package ffmpegconverter.converter;

import java.io.File;
import java.io.IOException;
import java.util.List;

public class AnimeConverter extends Converter {
    private String[] audioLang;
    private String[] subtitleLang;
    private boolean toStereo;
    private boolean preventSignsAndSongs;

    /**
     * @param ffprobePath           Path/command to ffprobe
     * @param ffmpegPath            Path/command to ffmpeg
     * @param audioLang             List of wanted audio languages in descending order
     * @param subtitleLang          List of wanted subtitle languages in descending order
     * @param toStereo              Convert video with several audio channels to stereo
     * @param preventSignsAndSongs  Prevent subtitles only converting signs and songs (not speech)
     */
    public AnimeConverter(String ffprobePath, String ffmpegPath, String[] audioLang, String[] subtitleLang, boolean toStereo, boolean preventSignsAndSongs) {
        this.ffprobePath = ffprobePath;
        this.ffmpegPath = ffmpegPath;
        this.audioLang = audioLang;
        this.subtitleLang = subtitleLang;
        this.toStereo = toStereo;
        this.preventSignsAndSongs = preventSignsAndSongs;
    }

    public void convert(File file) throws IOException {
        processFile(file.getParentFile(), file);
    }

    /**
     * Reads streams from a file, and converts it to an mp4.
     *
     * @param folder        The folder of the file to process
     * @param file          The file to process
     * @throws IOException  If the BufferedReader fails
     */
    private void processFile(File folder, File file) throws IOException {
        String[] streams = probeFile(ffprobePath, file);
        if (streams.length == 0) {
            throw new IllegalArgumentException("The file has no streams");
        }
        String noExt = file.getName().substring(0, file.getName().lastIndexOf('.'));
        String newPath = noExt + ".mp4";
        boolean isMP4 = newPath.equals(file.getName());
        if (isMP4) {
            newPath = noExt + ".tmp.mp4";
        }
        convertProcess(new ProcessBuilder(builderCommand(ffmpegPath, file.getName(), streams, newPath)), folder);
        if (isMP4) {
            File oldFile = new File(newPath);
            if (!oldFile.renameTo(new File(noExt + ".mp4"))) {
                System.out.println("Failed to move converted file.");
            }
        }
    }

    /**
     * Generates a command for a ProcessBuilder.
     *
     * @param executable    The executable file for ffmpeg
     * @param fileName      The input file
     * @param streams       A list of ffprobe streams
     * @param outFile       The output file
     * @return              A list of commands
     */
    private String[] builderCommand(String executable, String fileName, String[] streams, String outFile) {
        List<String> command = ffmpegWebVideo(executable, fileName);
        for (String lang : audioLang) {
            List<Integer> audioStreams = listAudio(streams, lang);
            if (audioStreams.size() > 0) {
                command.add("-map");
                command.add("0:" + audioStreams.get(0));
                String[] channels = stringBetween(streams[audioStreams.get(0)], "channels=", " ");
                if (toStereo && channels.length > 0 && Integer.parseInt(channels[0]) > 2) {
                    command.add("-af");
                    command.add("pan=stereo|FL=FC+0.30*FL+0.30*BL|FR=FC+0.30*FR+0.30*BR");
                }
                break;
            } else {
                audioStreams = listAudio(streams);
                if (audioStreams.size() > 0) {
                    command.add("-map");
                    command.add("0:" + audioStreams.get(0));
                }
            }
        }
        addSubtitles(streams, listVideo(streams), command, fileName);
        command.add(outFile);
        return command.toArray(new String[0]);
    }

    private void addSubtitles(String[] streams, List<Integer> videoStreams, List<String> command, String fileName) {
        for (String lang : subtitleLang) {
            List<Integer> subtitleStreams = listSubtitles(streams, lang);
            if (subtitleStreams.size() > 0 && videoStreams.size() > 0 && isImageSub(streams, subtitleStreams.get(0))) {
                command.add("-filter_complex");
                command.add("[0:v:" + listVideo(streams).get(0) + "][0:" + subtitleStreams.get(0) + "]overlay");
                break;
            } else if (subtitleStreams.size() > 0) {
                if (videoStreams.size() > 0) {
                    command.add("-map");
                    command.add("0:" + listVideo(streams).get(0));
                }
                if (preventSignsAndSongs) {
                    subtitleStreams = listSubtitlesRelative(streams, lang, new String[]{"title=Signs"});
                } else {
                    subtitleStreams = listSubtitlesRelative(streams, lang);
                }
                if (subtitleStreams.size() > 0) {
                    command.add("-vf");
                    command.add("subtitles='" + fileName.replace("'", "\'") + "':si=" + subtitleStreams.get(0));
                }
                break;
            }
        }
    }

    @Override
    public String[] getValidFormats() {
        return VIDEO_FORMATS;
    }
}
